{-|
Module      : Gargantext.Core.Text.Corpus.API.Hal
Description : Pubmed API connection
Copyright   : (c) CNRS, 2017
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE ViewPatterns       #-}

module Gargantext.Core.Text.Corpus.API.Hal
    ( getC
    , get

    -- * Internal functions for testing
    , getHalQuery
    , HalQuery(..)
    , convertQuery
    )
    where

import Conduit
import Data.Either
import Data.Maybe
import Data.Monoid
import Data.Semigroup
import Data.Text (Text, pack, intercalate)
import Gargantext.Core (Lang(..))
import Gargantext.Core.Text.Corpus.Query as Corpus
import Gargantext.Core.Types (Term(..))
import Gargantext.Database.Admin.Types.Hyperdata (HyperdataDocument(..))
import Gargantext.Prelude
import Network.HTTP.Types.URI (EscapeItem(..), renderQueryPartialEscape)
import Servant.Client (ClientError)
import qualified Data.Text as Text
import qualified Data.Text.Encoding as TE
import qualified Gargantext.Core.Text.Corpus.Parsers.Date as Date
import qualified Gargantext.Defaults as Defaults
import qualified HAL            as HAL
import qualified HAL.Client     as HAL
import qualified HAL.Doc.Corpus as HAL

newtype HalQuery = HalQuery { _HalQuery :: [EscapeItem] }
  deriving stock (Show, Eq)
  deriving newtype (Semigroup, Monoid)

-- | Returns an /url encoded/ query ready to be sent to Hal.
getHalQuery :: HalQuery -> Text
getHalQuery (HalQuery items) =
  Text.replace "q=" "" . TE.decodeUtf8 . renderQueryPartialEscape False $ [
    ("q", items)
    ]

convertQuery :: Corpus.Query -> HalQuery
convertQuery q = HalQuery (interpretQuery q transformAST)
  where

    lParen :: [EscapeItem]
    lParen = [QN "("]

    rParen :: [EscapeItem]
    rParen = [QN ")"]

    transformAST :: BoolExpr Term -> [EscapeItem]
    transformAST ast = case ast of
      BAnd sub1 sub2
        -> lParen <> transformAST sub1 <> [QN "+AND+"] <> transformAST sub2 <> rParen
      BOr sub1 sub2
        -> lParen <> transformAST sub1 <> [QN "+OR+"] <> transformAST sub2 <> rParen
      BNot (BConst (Negative term))
        -> transformAST (BConst (Positive term)) -- double negation
      BNot sub
        -> [QN "NOT+"] <> lParen <> transformAST sub <> rParen
      -- BTrue cannot happen is the query parser doesn't support parsing 'TRUE' alone.
      BTrue
        -> mempty
      -- BTrue cannot happen is the query parser doesn't support parsing 'FALSE' alone.
      BFalse
        -> mempty
      -- Maps the positive terms to contain 'abstract:'
      BConst (Positive (Term term))
        -> [QE (TE.encodeUtf8 term)]
      BConst (Negative sub)
        -> [QN "NOT+"] <> transformAST (BConst (Positive sub))

get :: Lang
    -> Corpus.Query
    -> Maybe Corpus.Limit
    -> IO [HyperdataDocument]
get la (convertQuery -> q) ml = do
  eDocs <- HAL.getMetadataWith (getHalQuery q) (Just 0) (fromIntegral . getLimit <$> ml)
  either (panic . pack . show) (\d -> mapM (toDoc' la) $ HAL._docs d) eDocs

halOptions :: HAL.HalCrawlerOptions
halOptions = HAL.HalCrawlerOptions False 1000

getC :: Lang
     -> Corpus.Query
     -> Maybe Corpus.Limit
     -> IO (Either ClientError (Maybe Integer, ConduitT () HyperdataDocument IO ()))
getC la (convertQuery -> q) ml = do
  eRes <- HAL.getMetadataWithOptionsC halOptions (getHalQuery q) (Just 0) (getLimit <$> ml)
  pure $ (\(len, docsC) -> (fromIntegral <$> len, docsC .| mapMC (toDoc' la))) <$> eRes
--  case eRes of
--    Left err -> panic $ pack $ show err
--    Right (len, docsC) -> pure (len, docsC .| mapMC (toDoc' la))

toDoc' :: Lang -> HAL.Corpus -> IO HyperdataDocument
toDoc' la (HAL.Corpus i t ab d s aus affs struct_id) = do
  (utctime, (pub_year, pub_month, pub_day)) <-
        Date.dateSplit la (maybe (Just $ pack $ show Defaults.year) Just d)
  pure HyperdataDocument { _hd_bdd = Just "Hal"
                         , _hd_doi = Just $ pack $ show i
                         , _hd_url = Nothing
                         , _hd_uniqId = Nothing
                         , _hd_uniqIdBdd = Nothing
                         , _hd_page = Nothing
                         , _hd_title = Just $ intercalate " " t
                         , _hd_authors = Just $ foldl (\x y -> x <> ", " <> y) "" aus
                         , _hd_institutes = Just $ foldl (\x y -> x <> ", " <> y) "" $ affs <> map (cs . show) struct_id
                         , _hd_source = Just $ maybe "Nothing" identity s
                         , _hd_abstract = Just $ intercalate " " ab
                         , _hd_publication_date = fmap (pack . show) utctime
                         , _hd_publication_year = pub_year
                         , _hd_publication_month = pub_month
                         , _hd_publication_day = pub_day
                         , _hd_publication_hour = Nothing
                         , _hd_publication_minute = Nothing
                         , _hd_publication_second = Nothing
                         , _hd_language_iso2 = Just $ (pack . show) la }
